function rad = adjust_radii(D, rad, min_dist)
% Adjust radii such that the minimum distance between atoms is satisfied.
%
% Input:
%   D           - dictionary (normalized atoms)
%   rad         - cone radii (if scalar, then all atoms have the same radius)
%   min_dist    - minimum distance between cones
%
% Output:
%   rad         - adjusted radii

% BD 22.05.2023

tol = 1e-12;
s_list = check_cone_superposition(D, rad, min_dist);

i = size(s_list,1);
%options = optimset('TolX',1e-4);
while ~isempty(s_list) && i>0  % check list one by one
  i1 = s_list(1,1);
  i2 = s_list(1,2);    % indices of atoms that are too close
  r = sqrt(2-2*abs(D(:,i1)'*D(:,i2))) - min_dist+tol;   % desired distance between the two atoms
  t = rad(i2)/rad(i1);
%  rad(i1) = fzero(@(x) x*sqrt(1-x*x*t*t/4) + x*t*sqrt(1-x*x/4) - r, rad(i1)*r/(rad(i1)+rad(i2)), options);
  rad(i1) = my_bisection(t, r, 0, rad(i1));
  rad(i2) = rad(i1)*t;
  s_list = check_cone_superposition(D, rad, min_dist);
  i=i-1;
end

if ~isempty(s_list)
  s_list
end

end

function a = my_bisection(t, r, a, b)
  f = @(x) x*sqrt(1-x*x*t*t/4) + x*t*sqrt(1-x*x/4) - r;
  tol = 1e-6;
  if 2/t < b  % adjust upper bound to keep the function real
    b = 2/t;
  end
  c = (a+b)/2;
  fc = f(c);
  while abs(fc) > tol
    c = (a+b)/2;
    fc = f(c);
    if f(c)<0
      a=c;
    else
      b=c;
    end
  end
end
