function s_list = check_cone_superposition(D, rad, min_dist)
% Check whether cone atoms have intersections
%
% Input:
%   D           - dictionary (normalized atoms)
%   rad         - cone radii (if scalar, then all atoms have the same radius)
%   min_dist    - minimum distance between cones (default = 0)
%
% Output:
%   s_list      - matrix with 2 columns; on each row, the indices of superposing atoms

% BD 31.07.2022

if nargin < 3
  min_dist = 0;
end

tol = 1e-12;

[~,n] = size(D);
if length (rad) == 1
  rad = rad * ones(n,1);
end
if length (rad) ~= n
  error('Radii vector has wrong length')
end
rad = rad(:);

G = eye(n);
for i = 1 : n
  for j = 1 : i
%     if D(:,i)'*D(:,j) < 0
%       G(i,j) = norm(D(:,i) + D(:,j));
%     else
%       G(i,j) = norm(D(:,i) - D(:,j));
%     end
    G(i,j) = sqrt(2 - 2*abs(D(:,i)'*D(:,j)));
  end
end
R = rad * sqrt(1 - rad'.*rad'/4) + sqrt(1 - rad.*rad/4) * rad' + min_dist;

[rl, cl] = find(tril(G,-1) + tol < tril(R,-1));
s_list = [rl, cl];