% test several DL algorithms on PYOD benchmark datasets

% BD 16.05.2023

%% ----------------------------------------
data_file_path = '../DATA_DEPENDENCY/dependency_';
% data_file_names = {'musk', 'SpamBase', 'landsat', 'satellite', 'satimage', ...
%                    'ionosphere', 'WPBC', 'letter', 'WDBC', 'fault', ...
%                    'cardio', 'Waveform', 'Hepatitis', 'Lymphography', 'pendigits', ...
%                    'wine', 'vowels', 'cover', 'magic.gamma.mat', 'breastw', ...
%                    'Stamps', 'WBC', 'yeast', 'glass', 'annthyroid', ...
%                    'mammography', 'Pima', 'thyroid', 'vertebral', 'Wilt'};
data_file_names = {'landsat', 'cardio'}; % uncomment full list above for more datasets
results_file_name = ['res_all_odds_', int2str(randi(9999)), '.csv'];

n_tests = 2; % 10
train_data_ratio = 0.7;

overcompleteness_list = [2 3]; % overcompleteness factor of the dictionary
s_list = [2];

rad_max = 0.05; %0.1;
rad_min = 0.05; %0.01;
iternum = 100;
iter_cd = 10;
err_thresh = 1e-7;
min_dist_cones = 0.01;

if rad_min ~= rad_max
  swap_rad = 1;
  method_names = {'AK-SVD + OMP', 'Cone-DL', 'AK-SVD + Cone-OMP', 'Cone-DL + swap', 'AK-SVD + Cone-OMP + swap'};
else
  swap_rad = 0;
  method_names = {'AK-SVD + OMP', 'Cone-DL', 'AK-SVD + Cone-OMP'};
end

% ----------------------------------------
n_datasets = length(data_file_names);
n_methods = length(method_names);
n_params = length(overcompleteness_list)*length(s_list);

% prepare file for results; write header
fid_res = fopen(results_file_name,'w');
fprintf(fid_res, "%s,%s,", 'Dataset', 'Method');
for i_over = 1:length(overcompleteness_list)
  for i_s = 1 : length(s_list)
    fprintf(fid_res, "%s,", ['o=',num2str(overcompleteness_list(i_over)), ' s=', int2str(s_list(i_s))]);
  end
end
fprintf(fid_res, "\n");

for i=1:n_methods
  res_methods{i} = zeros(n_datasets,n_params);
end

for id = 1 : n_datasets
  fprintf("%s\n", data_file_names{id})
  filename = [data_file_path, data_file_names{id}];
  load(filename)
  X = double(Y);   % data are in X, labels are in y
  y = labels;
  clear Y, clear labels
  X = X';   % signals are on columns
  [m,N] = size(X);

  % preprocessing: Z-score normalization
  vm = mean(X,2);
  X = double(X) - repmat(vm,1,N);         % subtract mean
  sm = std(X,0,2);
  X = X ./ repmat(sm,1,N);
  
  col_normal = find(y==0);
  col_anomaly = find(y~=0);
  Nanomaly = length(col_anomaly);
  Nnormal = N - Nanomaly;
  anomaly_ratio = Nanomaly/N;
  
  Ntrain = round(train_data_ratio*N);
  Ntest = N - Ntrain;
  N_anomaly_train = round(anomaly_ratio*Ntrain);
  N_normal_train = Ntrain - N_anomaly_train;
  N_anomaly_test = Nanomaly - N_anomaly_train;
  N_normal_test = Nnormal - N_normal_train;
  
  for im = 1:n_methods
    res_data{im} = zeros(n_tests, n_params);
    res_err{im} = zeros(n_tests, n_params);
  end
  
  for it = 1 : n_tests
    % split train and test data, respecting proportion of anomaly
    p1 = randperm(Nnormal);
    p2 = randperm(Nanomaly);
    Ytrain = X(:,[col_normal(p1(1:N_normal_train)); col_anomaly(p2(1:N_anomaly_train))]);
    Ytest = X(:,[col_normal(p1(N_normal_train+1:end)); col_anomaly(p2(N_anomaly_train+1:end))]);
    ytest = [zeros(N_normal_test,1); y(col_anomaly(p2(N_anomaly_train+1:end)))];
    N_anomaly_test_w = sum(ytest);
    
    i_param = 0;
      
    for i_over = 1:length(overcompleteness_list)
      n = round(m*overcompleteness_list(i_over));
      rad_unif = rad_min + (rad_max-rad_min) * rand(n,1);
      dist_ok = 0;
      i0 = 0;
      while ~dist_ok
        D0 = normc(randn(m,n));
        i1 = check_cone_superposition(D0, rad_unif, min_dist_cones);
        dist_ok = isempty(i1);
        i0 = i0+1;
        if i0 == 10 && dist_ok == 0
          %error('Cannot generate dictionary with imposed distance')
          rad_unif = 0.95*rad_unif;
          i0=0;
        end
      end
      %save('D0last', 'D0', 'rad_unif');
      %max(rad_unif)
      
      for i_s = 1 : length(s_list)
        i_param = i_param+1;
        s = s_list(i_s);
        rad = rad_unif;
            
        % training phase
        % standard DL (AK-SVD)
        D1 = DL(Ytrain, D0, s, iternum);

        % Cone-DL
        D2 = dl_cone_paksvd_opt(Ytrain, D0, s, iternum, rad, iter_cd, err_thresh, min_dist_cones);

        % testing phase
        % AK-SVD + OMP
        Xomp = omp(Ytest, D1, s);
        err = Ytest - D1*Xomp;
        err = sqrt(sum(err.*err));
        res_err{1}(it,i_param) = mean(err);
        [~,~,~,roc_auc] = perfcurve(ytest,err,1);
        res_data{1}(it,i_param) = roc_auc;

        % AK-SVD + Cone-OMP
        rad1 = adjust_radii(D1, rad, min_dist_cones);  % ensure that cones are disjoint
        err = zeros(Ntest,1);
        for k = 1 : Ntest
          [x, Da, support] = omp_cone_single(Ytest(:,k), D1, s, rad1, iter_cd, err_thresh);
          err(k) = norm(Ytest(:,k) - Da*x);
        end
        res_err{3}(it,i_param) = mean(err);
        [~,~,~,roc_auc] = perfcurve(ytest,err,1);
        res_data{3}(it,i_param) = roc_auc;

        % AK-SVD + Cone-OMP + swap
        if swap_rad
          % swap radii
          [~, Xs] = compute_rep_cone(D1, Ytrain, s, rad1, iter_cd, err_thresh);
          atom_use = sum(squeeze(Xs)~=0,2);
          [~,Ip] = sort(atom_use,'descend');
          [~,Ir] = sort(rad,'descend');
          rad_swap = zeros(n,1);
          rad_swap(Ip) = rad(Ir);
          rad_swap = adjust_radii(D1, rad_swap, min_dist_cones);

          % compute representation error
          err = zeros(Ntest,1);
          for k = 1 : Ntest
            [x, Da, support] = omp_cone_single(Ytest(:,k), D1, s, rad_swap, iter_cd, err_thresh);
            err(k) = norm(Ytest(:,k) - Da*x);
          end
          res_err{5}(it,i_param) = mean(err);
          [~,~,~,roc_auc] = perfcurve(ytest,err,1);
          res_data{5}(it,i_param) = roc_auc;
        end

        % Cone-DL
        err = zeros(Ntest,1);
        for k = 1 : Ntest
          [x, Da, support] = omp_cone_single(Ytest(:,k), D2, s, rad, iter_cd, err_thresh);
          err(k) = norm(Ytest(:,k) - Da*x);
        end
        res_err{2}(it,i_param) = mean(err);
        [~,~,~,roc_auc] = perfcurve(ytest,err,1);
        res_data{2}(it,i_param) = roc_auc;

        % Cone-DL + swap
        if swap_rad
          % swap radii
          [~, Xs] = compute_rep_cone(D2, Ytrain, s, rad, iter_cd, err_thresh);
          atom_use = sum(squeeze(Xs)~=0,2);
          [~,Ip] = sort(atom_use,'descend');
          [~,Ir] = sort(rad,'descend');
          rad_swap = zeros(n,1);
          rad_swap(Ip) = rad(Ir);
          rad_swap = adjust_radii(D2, rad_swap, min_dist_cones);

          % compute representation error
          err = zeros(Ntest,1);
          for k = 1 : Ntest
            [x, Da, support] = omp_cone_single(Ytest(:,k), D2, s, rad_swap, iter_cd, err_thresh);
            err(k) = norm(Ytest(:,k) - Da*x);
          end
          res_err{4}(it,i_param) = mean(err);
          [~,~,~,roc_auc] = perfcurve(ytest,err,1);
          res_data{4}(it,i_param) = roc_auc;
        end
      end
    end
  end
  %fprintf("ROC AUC: %f\n", mean(res_roc_auc(:,id)))
  for im = 1 : n_methods
    res_methods{im}(id,:) = mean(res_data{im});
    fprintf("%40s: %4f\n", method_names{im}, max(res_methods{im}(id,:)))
  end
  % write in file
  for im = 1 : n_methods
    fprintf(fid_res, "%s,%s,", data_file_names{id}, method_names{im});
    for ip = 1 : n_params
      fprintf(fid_res, "%f,", res_methods{im}(id,ip));
    end
    fprintf(fid_res, "\n");
  end
end

fclose(fid_res);
